import numpy as np
import re
from importlib.resources import open_binary
from crisprbact.utils import rev_comp, NoRecordsException
from crisprbact.off_target import (
    compute_off_target_df,
    extract_records,
    extract_features,
)

with open_binary("crisprbact", "reg_coef.pkl") as handle:
    coef = np.load(handle, allow_pickle=True)

bases = ["A", "T", "G", "C"]
GUIDE_LEN = 20


def encode(seq):
    """One-hot encoding of a sequence (only non-ambiguous bases (ATGC) accepted)"""

    return np.array([[int(b == p) for b in seq] for p in bases])


# Quartiles: q1 > 0.4 > q2 > -0.08 > q3 > -0.59 > q4
def predict(X):
    return [np.sum(x * coef) for x in X]


def find_targets(seq):
    repam = "[ATGC]GG"
    L = len(seq)
    seq_revcomp = rev_comp(seq)
    matching_targets = re.finditer(
        "(?=([ATGC]{6}" + repam + "[ATGC]{16}))", seq_revcomp
    )
    for target in matching_targets:
        matching_target = target.group(1)
        start, end = target.span(1)
        start_min = 14
        if start >= start_min:
            guide_start = start - start_min
            guide_end = end - 16 - 3
            guide = seq_revcomp[guide_start:guide_end]
            assert len(guide) == GUIDE_LEN
            pos_seq_start = L - guide_end
            pos_seq_stop = L - guide_start
            pos_seq_pam = pos_seq_start - 3
            yield dict(
                [
                    ("target", matching_target),
                    # ("guide", matching_target[:20]),
                    ("guide", guide),
                    ("start", pos_seq_start),
                    ("stop", pos_seq_stop),
                    ("pam", pos_seq_pam),
                    ("ori", "-"),
                ]
            )


def get_strand_value(value):
    strand_dict = {"+": 1, "1": 1, "-": -1, "-1": -1}
    return strand_dict[str(value)]


def on_target_predict(seq, genome=None, seed_sizes=[8, 9, 10, 11, 12, GUIDE_LEN]):

    seq = seq.upper()  # make uppercase
    seq = re.sub(r"\s", "", seq)  # removes white space
    records = None
    genome_features = None
    if genome:
        records = extract_records(genome)
        if records is None:
            raise NoRecordsException(
                "No records found in sequence file. Check the sequence or the format"
            )
        else:
            genome_features = extract_features(records)
    alltargets = list(find_targets(seq))
    if alltargets:
        X = np.array(
            [
                encode(target["target"][:7] + target["target"][9:])
                for target in alltargets
            ]  # encode and remove GG of PAM
        )
        X = X.reshape(X.shape[0], -1)
        preds = predict(X)
        for i, target in enumerate(alltargets):
            target_id = i + 1
            target.update({"target_id": target_id})
            target.update({"pred": preds[i]})
            if genome:
                off_targets_per_seed = []
                for seed_size in seed_sizes:
                    # off-target found for a guide
                    off_target_df = compute_off_target_df(
                        target["guide"], seed_size, records, genome_features
                    )
                    off_targets_list = []
                    if off_target_df is not None and not off_target_df.empty:
                        off_targets = off_target_df.loc[
                            0:,
                            [
                                "start",
                                "end",
                                "pampos",
                                "strand",
                                "recid",
                                "longest_perfect_match",
                                "pam_seq",
                                "features",
                            ],
                        ]
                        for j, off_t in enumerate(off_targets.values.tolist()):
                            off_target_dict = {
                                "off_target_id": str(target_id)
                                + "-"
                                + str(seed_size)
                                + "-"
                                + str(j),
                                "off_target_start": off_t[0],
                                "off_target_end": off_t[1],
                                "off_target_pampos": off_t[2],
                                "off_target_strand": off_t[3],
                                "off_target_recid": off_t[4],
                                "off_target_longest_perfect_match": off_t[5],
                                "off_target_pam_seq": off_t[6],
                                "off_target_good_orientation": None,
                            }
                            index_features = 7
                            # Loop through features associated to an off-target position
                            if len(off_t[index_features]) > 0:
                                # Loop for each feature
                                for feat in off_t[index_features]:
                                    if get_strand_value(
                                        off_target_dict["off_target_strand"]
                                    ) != get_strand_value(feat.location.strand):
                                        off_target_dict[
                                            "off_target_good_orientation"
                                        ] = True
                                    else:
                                        off_target_dict[
                                            "off_target_good_orientation"
                                        ] = False

                                    feature_dict = {
                                        "off_target_feat_strand": feat.location.strand,
                                        "off_target_feat_start": feat.location.start,
                                        "off_target_feat_end": feat.location.end,
                                        "off_target_feat_type": feat.type,
                                    }
                                    for k, feat in feat.qualifiers.items():
                                        if k != "translation":
                                            feature_dict["off_target_" + k] = "::".join(
                                                feat
                                            )
                                    off_targets_list.append(
                                        {**feature_dict, **off_target_dict}
                                    )
                            else:
                                off_targets_list.append(off_target_dict)
                        off_targets_per_seed.append(
                            {
                                "id": str(i) + "-" + str(seed_size),
                                "seed_size": seed_size,
                                "off_targets": off_targets_list,
                            }
                        )
                    else:
                        off_targets_per_seed.append(
                            {
                                "id": str(i) + "-" + str(seed_size),
                                "seed_size": seed_size,
                                "off_targets": [],
                            }
                        )
                target.update({"off_targets_per_seed": off_targets_per_seed})
            else:
                target.update({"off_targets_per_seed": []})
        return alltargets
    else:
        return []



import re
import pandas as pd

from crisprbact.utils import rev_comp
from itertools import chain


def compute_off_target_df(guide, seed_size, records, feature_df):
    """ Returns a pandas DataFrame with data about the identified off-targets.
    The features column contains a list of biopython SeqFeature objects that overlap
    with the off-target"""
    offs_df = get_off_target_pos(guide, records, seed_size)
    if offs_df is not None:
        offs_df["features"] = [
            get_pos_features(off.recid, off.pampos, feature_df)
            for i, off in offs_df.iterrows()
        ]
        return offs_df
    else:
        return None


def get_off_target_pos(guide, records, seed_size):
    if records is not None:
        offs = iter([])
        for rec in records:
            sequence_str = str(rec.seq).upper()
            seqid = rec.id
            offs_plus = re.finditer(guide[-seed_size:] + "[ATGC]GG", sequence_str)
            offs_plus = gen_extract_off_target_strand_plus(
                offs_plus, seqid, sequence_str, guide, seed_size
            )

            # - ori
            offs_minus = re.finditer(
                "CC[ATGC]" + rev_comp(guide[-seed_size:]), sequence_str
            )
            offs_minus = gen_extract_off_target_strand_minus(
                offs_minus, seqid, sequence_str, guide, seed_size
            )
            offs = chain(offs, offs_plus, offs_minus)
        offs_dict = dict(
            zip(
                [
                    "start",
                    "end",
                    "pampos",
                    "strand",
                    "recid",
                    "longest_perfect_match",
                    "pam_seq",
                ],
                zip(*offs),
            )
        )
        return pd.DataFrame(offs_dict)
    else:
        return None


def get_pos_features(recid, position, f_df):
    if len(f_df) > 0:
        feature_at_pos = f_df[
            (f_df.start < position) & (f_df.end > position) & (f_df.recid == recid)
        ]
        return feature_at_pos.feature.values
    else:
        return []


def gen_extract_off_target_strand_plus(
    off_target_matches, seqid, sequence_string, guide, seed_size
):
    for match in off_target_matches:
        guide_subseq = guide[: 20 - seed_size][::-1]
        # extract part of the sequence
        seq_extension_len = len(guide) - seed_size
        start_pos_seq = match.start() - seq_extension_len
        end_pos_seq = match.end() - seed_size - 3
        sub_sequence_to_match = sequence_string[start_pos_seq:end_pos_seq][::-1]
        matching_chars = list(common_start(sub_sequence_to_match, guide_subseq))
        matching_substr = "".join(matching_chars)
        longest_perfect_match = matching_substr[::-1] + match.group(0)[:-3]
        yield match.span() + (
            match.end(),
            "+",
            seqid,
            longest_perfect_match,
            match.group(0)[-3:],
        )


def common_start(seq1, seq2):
    for a, b in zip(seq1, seq2):
        if a == b:
            yield a
        else:
            return


def gen_extract_off_target_strand_minus(
    off_target_matches, seqid, sequence_string, guide, seed_size
):
    for match in off_target_matches:
        # Extract the sequence.
        # since rev_comp, extend the sequence to the end
        seq_extension_len = len(guide) - seed_size
        end_pos_seq = match.end() + seq_extension_len
        start_pos_seq = match.start() + seed_size + 3
        sub_sequence_to_match = sequence_string[start_pos_seq:end_pos_seq]

        # Extract part of the guide to match
        guide_subseq = rev_comp(guide[: 20 - seed_size])
        matching_chars = list(common_start(sub_sequence_to_match, guide_subseq))
        matching_substr = "".join(matching_chars)
        longest_perfect_match = match.group(0)[3:] + matching_substr

        yield match.span() + (
            match.start(),
            "-",
            seqid,
            longest_perfect_match,
            match.group(0)[:3],
        )


def extract_records(genome):
    records = list(genome)
    if records and len(records) > 0:
        return records
    else:
        return None


def extract_features(recs):

    if recs is not None:
        f_list = []
        for rec in recs:
            for f in rec.features:
                if f.type in ["CDS", "ncRNA", "rRNA", "tRNA"]:
                    f_list.append(
                        (
                            f.location.start.position,
                            f.location.end.position,
                            f.location.strand,
                            f.type,
                            f,
                            rec.id,
                        )
                    )
        f_dict = dict(
            zip(
                ["start", "end", "strand", "type", "feature", "recid"],
                zip(*f_list[1:]),
            )
        )  # starts at 1 to get rid of the first feature which is the whole chromosome
        return pd.DataFrame(f_dict)
    else:
        return None



from crisprbact.predict import on_target_predict
from crisprbact.off_target import (
    compute_off_target_df,
    extract_features,
    extract_records,
)
from crisprbact.utils import NoRecordsException

__all__ = [
    "extract_records",
    "on_target_predict",
    "compute_off_target_df",
    "extract_features",
    "NoRecordsException",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crisprbact']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.75,<2.0',
 'click>=7.0,<8.0',
 'numpy>=1.17,<2.0',
 'pandas>=1.0.0,<2.0.0',
 'rope>=0.16.0,<0.17.0']

entry_points = \
{'console_scripts': ['crisprbact = crisprbact.cli:main']}

setup_kwargs = {
    'name': 'crisprbact',
    'version': '0.3.10',
    'description': 'Tools to design and analyse CRISPRi experiments',
    'long_description': '# CRISPRbact\n\n**Tools to design and analyse CRISPRi experiments in bacteria.**\n\nCRISPRbact currently contains an on-target activity prediction tool for the Streptococcus pyogenes dCas9 protein.\nThis tool takes as an input the sequence of a gene of interest and returns a list of possible target sequences with the predicted on-target activity. Predictions are made using a linear model fitted on data from a genome-wide CRISPRi screen performed in E. coli (Cui et al. Nature Communications, 2018). The model predicts the ability of dCas9 to block the RNA polymerase when targeting the non-template strand (i.e. the coding strand) of a target gene.\n\n## Getting Started\n\n### Installation\n\nFor the moment, you can install this package only via PyPI\n\n#### PyPI\n\n```console\n$ pip install crisprbact\n$ crisprbact --help\n```\n\n```\nUsage: crisprbact [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -v, --verbose\n  --help         Show this message and exit.\n\nCommands:\n  predict\n```\n\n### API\n\nUsing this library in your python code.\n\n```python\nfrom crisprbact import on_target_predict\n\nguide_rnas = on_target_predict("ACCACTGGCGTGCGCGTTACTCATCAGATGCTGTTCAATACCGATCAGGTTATCGAAGTGTTTGTGATTGTTTGCCGCGCGCGTGGCGAAGGCCCGTGATGAAGGAAAAGTTTTGCGCTATGTTGGCAATATTGATGAAG")\n\nfor guide_rna in guide_rnas:\n    print(guide_rna)\n\n```\n\n_output :_\n\n```\n{\'target\': \'TCATCACGGGCCTTCGCCACGCGCG\', \'guide\': \'TCATCACGGGCCTTCGCCAC\', \'start\': 82, \'stop\': 102, \'pam\': 80, \'ori\': \'-\', \'target_id\': 1, \'pred\': -0.4719254873780802, \'off_targets_per_seed\': []}\n{\'target\': \'CATCACGGGCCTTCGCCACGCGCGC\', \'guide\': \'CATCACGGGCCTTCGCCACG\', \'start\': 81, \'stop\': 101, \'pam\': 79, \'ori\': \'-\', \'target_id\': 2, \'pred\': 1.0491308060379676, \'off_targets_per_seed\': []}\n{\'target\': \'CGCGCGCGGCAAACAATCACAAACA\', \'guide\': \'CGCGCGCGGCAAACAATCAC\', \'start\': 63, \'stop\': 83, \'pam\': 61, \'ori\': \'-\', \'target_id\': 3, \'pred\': -0.9021152826078697, \'off_targets_per_seed\': []}\n{\'target\': \'CCTGATCGGTATTGAACAGCATCTG\', \'guide\': \'CCTGATCGGTATTGAACAGC\', \'start\': 29, \'stop\': 49, \'pam\': 27, \'ori\': \'-\', \'target_id\': 4, \'pred\': 0.23853258873311955, \'off_targets_per_seed\': []}\n```\n\n### Command line interface\n\n#### Predict guide RNAs activity\n\nInput the sequence of a target gene and this script will output candidate guide RNAs for the S. pyogenes dCas9 with predicted on-target activity.\n\n```console\n$ crisprbact predict --help\n```\n\n```\nUsage: crisprbact predict [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  from-seq  Outputs candidate guide RNAs for the S.\n  from-str  Outputs candidate guide RNAs for the S.\n```\n\n##### From a string sequence:\n\nThe target input sequence can be a simple string.\n\n```console\n$ crisprbact predict from-str --help\n```\n\n```\nUsage: cli.py predict from-str [OPTIONS] [OUTPUT_FILE]\n\n  Outputs candidate guide RNAs for the S. pyogenes dCas9 with predicted on-\n  target activity from a target gene.\n\n  [OUTPUT_FILE] file where the candidate guide RNAs are saved. Default =\n  "stdout"\n\nOptions:\n  -t, --target TEXT               Sequence file to target  [required]\n  -s, --off-target-sequence FILENAME\n                                  Sequence in which you want to find off-\n                                  targets\n  -w, --off-target-sequence-format [fasta|gb|genbank]\n                                  Sequence in which you want to find off-\n                                  targets format  [default: genbank]\n  --help                          Show this message and exit.\n\n```\n\n```console\n$ crisprbact predict from-str -t ACCACTGGCGTGCGCGTTACTCATCAGATGCTGTTCAATACCGATCAGGTTATCGAAGTGTTTGTGATTGTTTGCCGCGCGCGTGGCGAAGGCCCGTGATGAAGGAAAAGTTTTGCGCTATGTTGGCAATATTGATGAAG guide-rnas.tsv\n```\n\noutput file `guide-rnas.tsv` :\n\nNo `seq_id` is defined since it is from a simple string.\n\n```\ntarget\tPAM position\tprediction\tseq_id\nTCATCACGGGCCTTCGCCACGCGCG\t80\t-0.4719254873780802\tN/A\nCATCACGGGCCTTCGCCACGCGCGC\t79\t1.0491308060379676\tN/A\nCGCGCGCGGCAAACAATCACAAACA\t61\t-0.9021152826078697\tN/A\nCCTGATCGGTATTGAACAGCATCTG\t27\t0.23853258873311955\tN/A\n```\n\nYou can also pipe the results :\n\n```console\n$ crisprbact predict from-str -t ACCACTGGCGTGCGCGTTACTCATCAGATGCTGTTCAATACCGATCAGGTTATCGAAGTGTTTGTGATTGTTTGCCGCGCGCGTGGCGAAGGCCCGTGATGAAGGAAAAGTTTTGCGCTATGTTGGCAATATTGATGAAG | tail -n +2 | wc -l\n```\n\n##### From a sequence file\n\n```console\n$ crisprbact predict from-seq --help\n```\n\n```\nUsage: cli.py predict from-seq [OPTIONS] [OUTPUT_FILE]\n\n  Outputs candidate guide RNAs for the S. pyogenes dCas9 with predicted on-\n  target activity from a target gene.\n\n  [OUTPUT_FILE] file where the candidate guide RNAs are saved. Default =\n  "stdout"\n\nOptions:\n  -t, --target FILENAME           Sequence file to target  [required]\n  -f, --seq-format [fasta|gb|genbank]\n                                  Sequence file to target format  [default:\n                                  fasta]\n  -s, --off-target-sequence FILENAME\n                                  Sequence in which you want to find off-\n                                  targets\n  -w, --off-target-sequence-format [fasta|gb|genbank]\n                                  Sequence in which you want to find off-\n                                  targets format  [default: genbank]\n  --help                          Show this message and exit.\n```\n\n- Fasta file (could be a multifasta file)\n\n```console\n$ crisprbact predict from-seq -t /tmp/seq.fasta guide-rnas.tsv\n```\n\n- GenBank file\n\n```console\n$ crisprbact predict from-seq -t /tmp/seq.gb -f gb guide-rnas.tsv\n```\n\n- Off-targets\n\n```console\npredict from-seq -t data-test/sequence.fasta -s data-test/sequence.gb guide-rnas.tsv\n```\n\n##### Output file\n\n```\ntarget_id\ttarget\tPAM position\tprediction\ttarget_seq_id\tseed_size\toff_target_recid\toff_target_start\toff_target_end\toff_target_pampos\toff_target_strand\toff_target_feat_type\toff_target_feat_start\toff_target_feat_end\toff_target_feat_strand\toff_target_locus_tag\toff_target_gene\toff_target_note\toff_target_product\toff_target_protein_id\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t1388198\t1388209\t1388209\t+\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t2244514\t2244525\t2244525\t+\tCDS\t2243562\t2244720\t-1\tNRG857_10810\t\tCOG1174 ABC-type proline/glycine betaine transport systems, permease component\tputative transport system permease\tYP_006120510.1\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t4160984\t4160995\t4160995\t+\tCDS\t4160074\t4161406\t-1\tNRG857_19625\thslU\tCOG1220 ATP-dependent protease HslVU (ClpYQ), ATPase subunit\tATP-dependent protease ATP-binding subunit HslU\tYP_006122267.1\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t4534189\t4534200\t4534200\t+\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t548804\t548815\t548804\t-\n1\tTGATCCAGGCATTTTTTAGCTTCAT\t835\t0.47949500169043713\tNC_017634.1:2547433-2548329\t8\tNC_017634.1\t786462\t786473\t786462\t-\tCDS\t785384\t786470\t1\tNRG857_03580\t\tCOG2055 Malate/L-lactate dehydrogenases\thypothetical protein\tYP_006119079.1\n\n```\n\n## Contributing\n\n### Clone repo\n\n```console\n$ git clone https://gitlab.pasteur.fr/dbikard/crisprbact.git\n```\n\n### Create a virtualenv\n\n```console\n$ virtualenv -p python3.7 .venv\n$ . .venv/bin/activate\n$ pip install poetry\n```\n\n### Install crisprbact dependencies\n\n```console\n$ poetry install\n```\n\n### Install hooks\n\nIn order to run flake8 and black for each commit.\n\n```console\n$ pre-commit install\n```\n',
    'author': 'David Bikard',
    'author_email': 'david.bikard@pasteur.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.pasteur.fr/dbikard/crisprbact',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from Bio import SeqIO
import click
from crisprbact.predict import on_target_predict


class Config(object):
    def __init__(self):
        self.verbose = False


pass_config = click.make_pass_decorator(Config, ensure=True)

OFF_TARGET_DETAILS = [
    "off_target_recid",
    "off_target_start",
    "off_target_end",
    "off_target_pampos",
    "off_target_strand",
    "off_target_perfect_match",
    "off_target_good_orientation",
    "off_target_feat_type",
    "off_target_feat_start",
    "off_target_feat_end",
    "off_target_feat_strand",
    "off_target_locus_tag",
    "off_target_gene",
    "off_target_note",
    "off_target_product",
    "off_target_protein_id",
]
HEADER = [
    "target_id",
    "guide",
    "guide_start",
    "guide_end",
    "PAM position",
    "prediction",
    "target_seq_id",
    "seed_size",
] + OFF_TARGET_DETAILS
SEED_SIZE = 8
GENOME_FORMAT = "genbank"


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@pass_config
def main(config, verbose):
    config.verbose = verbose


@main.group()
@pass_config
def predict(config):
    pass


@predict.command()
@click.option(
    "-t", "--target", type=str, required=True, help="Sequence file to target",
)
@click.option(
    "-s",
    "--off-target-sequence",
    type=click.File("rU"),
    help="Sequence in which you want to find off-targets",
)
@click.option(
    "-w",
    "--off-target-sequence-format",
    type=click.Choice(["fasta", "gb", "genbank"]),
    default=GENOME_FORMAT,
    show_default=True,
    help="Sequence in which you want to find off-targets format",
)
@click.argument("output-file", type=click.File("w"), default="-")
@pass_config
def from_str(
    config, target, off_target_sequence, off_target_sequence_format, output_file
):
    """
    Outputs candidate guide RNAs for the S. pyogenes dCas9 with predicted on-target
    activity from a target gene.

    [OUTPUT_FILE] file where the candidate guide RNAs are saved. Default = "stdout"

    """
    if config.verbose:
        print_parameters(target)
    if off_target_sequence:
        genome_fh = SeqIO.parse(off_target_sequence, off_target_sequence_format)
    else:
        genome_fh = None
    guide_rnas = on_target_predict(target, genome_fh)

    click.echo("\t".join(HEADER), file=output_file)
    write_guide_rnas(guide_rnas, output_file)


@predict.command()
@click.option(
    "-t",
    "--target",
    type=click.File("rU"),
    required=True,
    help="Sequence file to target",
)
@click.option(
    "-f",
    "--seq-format",
    type=click.Choice(["fasta", "gb", "genbank"]),
    help="Sequence file to target format",
    default="fasta",
    show_default=True,
)
# @click.option(
#     "-s", "--seed-size", type=click.IntRange(8, 15, clamp=True),
# )
@click.option(
    "-s",
    "--off-target-sequence",
    type=click.File("rU"),
    help="Sequence in which you want to find off-targets",
)
@click.option(
    "-w",
    "--off-target-sequence-format",
    type=click.Choice(["fasta", "gb", "genbank"]),
    default=GENOME_FORMAT,
    show_default=True,
    help="Sequence in which you want to find off-targets format",
)
@click.argument("output-file", type=click.File("w"), default="-")
@pass_config
def from_seq(
    config,
    target,
    seq_format,
    off_target_sequence,
    off_target_sequence_format,
    output_file,
):
    """
    Outputs candidate guide RNAs for the S. pyogenes dCas9 with predicted on-target
    activity from a target gene.

    [OUTPUT_FILE] file where the candidate guide RNAs are saved. Default = "stdout"

    """
    fg = "blue"
    if config.verbose:
        print_parameters(target.name, fg)
    click.echo("\t".join(HEADER), file=output_file)
    for record in SeqIO.parse(target, seq_format):
        if config.verbose:
            click.secho(" - search guide RNAs for %s " % record.id, fg=fg)
        if off_target_sequence:
            genome_fh = SeqIO.parse(off_target_sequence, off_target_sequence_format)
        else:
            genome_fh = None
        guide_rnas = on_target_predict(str(record.seq), genome_fh)
        # print(guide_rnas)
        write_guide_rnas(guide_rnas, output_file, record.id)


def print_parameters(target, fg="blue"):
    click.secho("[Verbose mode]", fg=fg)
    click.secho("Target sequence : %s" % target, fg=fg)


def write_guide_rnas(guide_rnas, output_file, seq_id="N/A"):
    for guide_rna in guide_rnas:
        row = [
            str(guide_rna["target_id"]),
            guide_rna["guide"],
            str(guide_rna["start"]),
            str(guide_rna["stop"]),
            str(guide_rna["pam"]),
            str(guide_rna["pred"]),
            seq_id,
        ]
        if len(guide_rna["off_targets_per_seed"]) > 0:
            for off_target_per_seed in guide_rna["off_targets_per_seed"]:
                for off_target in off_target_per_seed["off_targets"]:
                    seed_size = off_target_per_seed["seed_size"]

                    def extract_off_target_detail(key):
                        if key in off_target:
                            return str(off_target[key])
                        else:
                            return ""

                    details = map(extract_off_target_detail, OFF_TARGET_DETAILS)
                    click.echo(
                        "\t".join(row + [str(seed_size)] + list(details)),
                        file=output_file,
                    )
        else:
            click.echo(
                "\t".join(row + list(map(lambda x: "", OFF_TARGET_DETAILS))),
                file=output_file,
            )


if __name__ == "__main__":
    main()
